@@ -46,7 +46,7 @@ class PolynomialScaling(object):
     def from_properties(properties, scale_index):
         try:
             number_of_coefficients = properties[
-                'NI_Scale[%d]_Polynomial_Coefficients_Size' % (scale_index)]
+                'NI_Scale[%d]_Polynomial_Coefficients_Size' % scale_index]
         except KeyError:
             number_of_coefficients = 4
         try:
@@ -61,6 +61,11 @@ class PolynomialScaling(object):
         return PolynomialScaling(coefficients, input_source)
 
     def scale(self, data):
+        if len(self.coefficients) == 0:
+            return np.zeros(len(data), dtype=np.dtype('float64'))
+
+        # Ensure data is double type before scaling
+        data = data.astype(np.dtype('float64'), copy=False)
         return np.polynomial.polynomial.polyval(data, self.coefficients)
 
 
@@ -229,7 +234,7 @@ class ThermistorScaling(object):
         """ Convert voltage data to temperature in Kelvin
         """
         # Ensure data is double precision
-        data = data.astype(np.float64, copy=False)
+        data = data.astype(np.dtype('float64'), copy=False)
         if self.excitation_type == CURRENT_EXCITATION:
             r_t = data / self.excitation_value
         elif self.excitation_type == VOLTAGE_EXCITATION:
@@ -376,6 +381,26 @@ class MultiScaling(object):
         final_scale = len(self.scalings) - 1
         return self._compute_scaled_data(final_scale, raw_channel_data)
 
+    def get_dtype(self, raw_data_type, scaler_data_types):
+        """ Get the numpy dtype for scaled data
+        """
+        final_scale = len(self.scalings) - 1
+        return self._compute_scale_dtype(final_scale, raw_data_type, scaler_data_types)
+
+    def _compute_scale_dtype(self, scale_index, raw_data_type, scaler_data_types):
+        if scale_index == RAW_DATA_INPUT_SOURCE:
+            return raw_data_type.nptype
+        scaling = self.scalings[scale_index]
+        if isinstance(scaling, DaqMxScalerScaling):
+            return scaler_data_types[scaling.scale_id].nptype
+        elif isinstance(scaling, AddScaling) or isinstance(scaling, SubtractScaling):
+            return np.result_type(
+                self._compute_scale_dtype(scaling.left_input_source, raw_data_type, scaler_data_types),
+                self._compute_scale_dtype(scaling.right_input_source, raw_data_type, scaler_data_types))
+        else:
+            # Any other scaling type should produce double data
+            return np.dtype('float64')
+
     def _compute_scaled_data(self, scale_index, raw_channel_data):
         """ Compute output data from a single scale in the set of all scalings,
             computing any required input scales recursively.
@@ -424,6 +449,10 @@ def _get_channel_scaling(properties):
     num_scalings = _get_number_of_scalings(properties)
     if num_scalings is None or num_scalings == 0:
         return None
+    scaling_status = properties.get("NI_Scaling_Status", "unscaled")
+    if scaling_status == "scaled":
+        # Data is written with scaling already applied
+        return None
 
     scalings = [None] * num_scalings
     for scale_index in range(num_scalings):
