@@ -3,16 +3,18 @@
     This module contains the public facing API for reading TDMS files
 """
 
+from collections import defaultdict
 import warnings
 import numpy as np
 
-from nptdms import scaling
-from nptdms.utils import Timer, OrderedDict
+from nptdms import scaling, types
+from nptdms.utils import Timer, OrderedDict, cached_property
 from nptdms.log import log_manager
 from nptdms.common import ObjectPath
 from nptdms.reader import TdmsReader
 from nptdms.channel_data import get_data_receiver
 from nptdms.export import hdf_export, pandas_export
+from nptdms.base_segment import RawChannelDataChunk
 
 
 log = log_manager.get_logger(__name__)
@@ -26,18 +28,40 @@ _property_builtin = property
 class TdmsFile(object):
     """ Reads and stores data from a TDMS file.
 
-    Can be indexed by group name to access a group within the TDMS file, for example::
+    There are two main ways to create a new TdmsFile object.
+    TdmsFile.read will read all data into memory::
+
+        tdms_file = TdmsFile.read(tdms_file_path)
+
+    or you can use TdmsFile.open to read file metadata but not immediately read all data,
+    for cases where a file is too large to easily fit in memory or you don't need to
+    read data for all channels::
+
+        with TdmsFile.open(tdms_file_path) as tdms_file:
+            # Use tdms_file
+            ...
+
+    This class acts like a dictionary, where the keys are names of groups in the TDMS
+    files and the values are TdmsGroup objects.
+    A TdmsFile can be indexed by group name to access a group within the TDMS file, for example::
 
         tdms_file = TdmsFile.read(tdms_file_path)
         group = tdms_file[group_name]
+
+    Iterating over a TdmsFile produces the names of groups in this file,
+    or you can use the groups method to directly access all groups::
+
+        for group in tdms_file.groups():
+            # Use group
+            ...
     """
 
     @staticmethod
     def read(file, memmap_dir=None):
         """ Creates a new TdmsFile object and reads all data in the file
 
-        :param file: Either the path to the TDMS file to read or an already
-            opened file.
+        :param file: Either the path to the tdms file to read
+            as a string or pathlib.Path, or an already opened file.
         :param memmap_dir: The directory to store memory mapped data files in,
             or None to read data into memory. The data files are created
             as temporary files and are deleted when the channel data is no
@@ -51,8 +75,8 @@ class TdmsFile(object):
         """ Creates a new TdmsFile object and reads metadata, leaving the file open
             to allow reading channel data
 
-        :param file: Either the path to the TDMS file to read or an already
-            opened file.
+        :param file: Either the path to the tdms file to read
+            as a string or pathlib.Path, or an already opened file.
         :param memmap_dir: The directory to store memory mapped data files in,
             or None to read data into memory. The data files are created
             as temporary files and are deleted when the channel data is no
@@ -65,16 +89,16 @@ class TdmsFile(object):
     def read_metadata(file):
         """ Creates a new TdmsFile object and only reads the metadata
 
-        :param file: Either the path to the TDMS file to read or an already
-            opened file.
+        :param file: Either the path to the tdms file to read
+            as a string or pathlib.Path, or an already opened file.
         """
         return TdmsFile(file, read_metadata_only=True)
 
     def __init__(self, file, memmap_dir=None, read_metadata_only=False, keep_open=False):
         """Initialise a new TdmsFile object
 
-        :param file: Either the path to the TDMS file to read or an already
-            opened file.
+        :param file: Either the path to the tdms file to read
+            as a string or pathlib.Path, or an already opened file.
         :param memmap_dir: The directory to store memory mapped data files in,
             or None to read data into memory. The data files are created
             as temporary files and are deleted when the channel data is no
@@ -91,6 +115,7 @@ class TdmsFile(object):
         self._properties = {}
         self._channel_data = {}
         self._reader = None
+        self.data_read = False
 
         reader = TdmsReader(file)
         try:
@@ -144,6 +169,19 @@ class TdmsFile(object):
         """
         return hdf_export.from_tdms_file(self, filepath, mode, group)
 
+    def data_chunks(self):
+        """ A generator that streams chunks of data from disk.
+        This method may only be used when the TDMS file was opened without reading all data immediately.
+
+        :rtype: Generator that yields :class:`DataChunk` objects
+        """
+        reader = self._get_reader()
+        channel_offsets = defaultdict(int)
+        for chunk in reader.read_raw_data():
+            yield DataChunk(self, chunk, channel_offsets)
+            for path, data in chunk.channel_data.items():
+                channel_offsets[path] += len(data)
+
     def close(self):
         """ Close the underlying file if it was opened by this TdmsFile
 
@@ -154,6 +192,16 @@ class TdmsFile(object):
             self._reader.close()
             self._reader = None
 
+    def __len__(self):
+        """ Returns the number of groups in this file
+        """
+        return len(self._groups)
+
+    def __iter__(self):
+        """ Returns an iterator over the names of groups in this file
+        """
+        return iter(self._groups)
+
     def __getitem__(self, group_name):
         """ Retrieve a TDMS group from the file by name
         """
@@ -168,6 +216,12 @@ class TdmsFile(object):
     def __exit__(self, exc_type, exc_value, traceback):
         self.close()
 
+    def _get_reader(self):
+        if self._reader is None:
+            raise RuntimeError(
+                "Cannot read data after the underlying TDMS reader is closed")
+        return self._reader
+
     def _read_file(self, tdms_reader, read_metadata_only):
         tdms_reader.read_metadata()
 
@@ -219,14 +273,13 @@ class TdmsFile(object):
         with Timer(log, "Read data"):
             # Now actually read all the data
             for chunk in tdms_reader.read_raw_data():
-                for (path, data) in chunk.raw_data.items():
+                for (path, data) in chunk.channel_data.items():
                     channel_data = self._channel_data[path]
-                    channel_data.append_data(data)
-                for (path, data) in chunk.daqmx_raw_data.items():
-                    channel_data = self._channel_data[path]
-                    for scaler_id, scaler_data in data.items():
-                        channel_data.append_scaler_data(
-                            scaler_id, scaler_data)
+                    if data.data is not None:
+                        channel_data.append_data(data.data)
+                    elif data.scaler_data is not None:
+                        for scaler_id, scaler_data in data.scaler_data.items():
+                            channel_data.append_scaler_data(scaler_id, scaler_data)
 
             for group in self.groups():
                 for channel in group.channels():
@@ -234,16 +287,24 @@ class TdmsFile(object):
                     if channel_data is not None:
                         channel._set_raw_data(channel_data)
 
+        self.data_read = True
+
+    def _read_channel_data_chunks(self, channel):
+        reader = self._get_reader()
+        for chunk in reader.read_raw_data_for_channel(channel.path):
+            yield chunk
+
+    def _read_channel_data_chunk_for_index(self, channel, index):
+        return self._get_reader().read_channel_chunk_for_index(channel.path, index)
+
     def _read_channel_data(self, channel, offset=0, length=None):
         if offset < 0:
             raise ValueError("offset must be non-negative")
         if length is not None and length < 0:
             raise ValueError("length must be non-negative")
-        if self._reader is None:
-            raise RuntimeError(
-                "Cannot read channel data after the underlying TDMS reader is closed")
+        reader = self._get_reader()
 
-        with Timer(log, "Allocate space"):
+        with Timer(log, "Allocate space for channel"):
             # Allocate space for data
             if length is None:
                 num_values = len(channel) - offset
@@ -252,13 +313,13 @@ class TdmsFile(object):
             num_values = max(0, num_values)
             channel_data = get_data_receiver(channel, num_values, self._memmap_dir)
 
-        with Timer(log, "Read data"):
+        with Timer(log, "Read data for channel"):
             # Now actually read all the data
-            for chunk in self._reader.read_raw_data_for_channel(channel.path, offset, length):
-                if chunk.raw_data is not None:
-                    channel_data.append_data(chunk.raw_data)
-                if chunk.daqmx_raw_data is not None:
-                    for scaler_id, scaler_data in chunk.daqmx_raw_data.items():
+            for chunk in reader.read_raw_data_for_channel(channel.path, offset, length):
+                if chunk.data is not None:
+                    channel_data.append_data(chunk.data)
+                if chunk.scaler_data is not None:
+                    for scaler_id, scaler_data in chunk.scaler_data.items():
                         channel_data.append_scaler_data(scaler_id, scaler_data)
 
         return channel_data
@@ -266,10 +327,10 @@ class TdmsFile(object):
     def object(self, *path):
         """(Deprecated) Get a TDMS object from the file
 
-        :param path: The object group and channel. Providing no channel
+        :param path: The object group and channel names. Providing no channel
             returns a group object, and providing no channel or group
             will return the root object.
-        :rtype: :class:`TdmsObject`
+        :rtype: One of :class:`TdmsGroup`, :class:`TdmsChannel`, :class:`RootObject`
 
         For example, to get the root object::
 
@@ -289,6 +350,13 @@ class TdmsFile(object):
                     "TdmsFile[group_name] to access a group object and " +
                     "TdmsFile[group_name][channel_name] to access a channel object.")
 
+        def get_name(component):
+            try:
+                return component.name
+            except AttributeError:
+                return component
+
+        path = [get_name(c) for c in path]
         object_path = ObjectPath(*path)
         try:
             return self.objects[str(object_path)]
@@ -350,10 +418,19 @@ class TdmsFile(object):
 class TdmsGroup(object):
     """ Represents a group of channels in a TDMS file.
 
-    Can be indexed by channel name to access a channel in this group, for example::
+    This class acts like a dictionary, where the keys are names of channels in the group
+    and the values are TdmsChannel objects.
+    A TdmsGroup can be indexed by channel name to access a channel in this group, for example::
 
         channel = group[channel_name]
 
+    Iterating over a TdmsGroup produces the names of channels in this group,
+    or you can use the channels method to directly access all channels::
+
+        for channel in group.channels():
+            # Use channel
+            ...
+
     :ivar ~.properties: Dictionary of TDMS properties defined for this group.
     """
 
@@ -400,6 +477,16 @@ class TdmsGroup(object):
 
         return pandas_export.from_group(self, time_index, absolute_time, scaled_data)
 
+    def __len__(self):
+        """ Returns the number of channels in this group
+        """
+        return len(self._channels)
+
+    def __iter__(self):
+        """ Returns an iterator over the names of channels in this group
+        """
+        return iter(self._channels)
+
     def __getitem__(self, channel_name):
         """ Retrieve a TDMS channel from this group by name
         """
@@ -452,7 +539,23 @@ class TdmsGroup(object):
 class TdmsChannel(object):
     """ Represents a data channel in a TDMS file.
 
-    To find the number of data points in this channel use :code:`len(channel)`.
+    This class acts like an array, you can get the length of a channel using :code:`len(channel)`,
+    and can iterate over values in the channel using a for loop,
+    or index into a channel using an integer index to get a single value::
+
+        for value in channel:
+            # Use value
+            ...
+        first_value = channel[0]
+
+    Or you can index using a slice to retrieve a range of data as a numpy array.
+    To get all data in this channel as a numpy array::
+
+        all_data = channel[:]
+
+    Or to retrieve a subset of data::
+
+        data_subset = channel[start:stop]
 
     :ivar ~.properties: Dictionary of TDMS properties defined for this channel,
                       for example the start time and time increment for waveforms.
@@ -469,14 +572,37 @@ class TdmsChannel(object):
         self.scaler_data_types = scaler_data_types
 
         self._raw_data = None
-        self._data_scaled = None
+        self._cached_chunk = None
+        self._cached_chunk_bounds = None
 
     def __repr__(self):
         return "<TdmsChannel with path %s>" % self.path
 
     def __len__(self):
+        """ Returns the number of values in this channel
+        """
         return self._length
 
+    def __iter__(self):
+        """ Returns an iterator over the values in this channel
+        """
+        if self._raw_data is not None:
+            return iter(self.data)
+        else:
+            return self._read_data_values()
+
+    def __getitem__(self, index):
+        if self._raw_data is not None:
+            return self.data[index]
+        elif index is Ellipsis:
+            return self.read_data()
+        elif isinstance(index, slice):
+            return self._read_slice(index.start, index.stop, index.step)
+        elif isinstance(index, int):
+            return self._read_at_index(index)
+        else:
+            raise TypeError("Invalid index type '%s', expected int, slice or Ellipsis" % type(index).__name__)
+
     @_property_builtin
     def path(self):
         """ Path to the TDMS object for this channel
@@ -489,31 +615,55 @@ class TdmsChannel(object):
         """
         return self._path.channel
 
-    @_property_builtin
-    def data(self):
+    @cached_property
+    def dtype(self):
+        """ NumPy data type of the channel data
+
+        For data with a scaling this is the data type of the scaled data
+
+        :rtype: numpy.dtype
         """
-        NumPy array containing the data for this channel
+        channel_scaling = self._scaling
+        if channel_scaling is not None:
+            return channel_scaling.get_dtype(self.data_type, self.scaler_data_types)
+        return self._raw_data_dtype()
+
+    def _raw_data_dtype(self):
+        if self.data_type is types.String:
+            return np.dtype('O')
+        elif self.data_type is types.TimeStamp:
+            return np.dtype('<M8[us]')
+        if self.data_type is not None and self.data_type.nptype is not None:
+            return self.data_type.nptype
+        return np.dtype('V8')
+
+    @cached_property
+    def data(self):
+        """ If the TdmsFile was created by reading all data, this property
+        provides direct access to the numpy array containing the data for this channel.
+
+        Indexing into the channel with a slice should be preferred to using this property, for example::
+
+            channel_data = channel[:]
         """
         if len(self) > 0 and self._raw_data is None:
             raise RuntimeError("Channel data has not been read")
 
         if self._raw_data is None:
-            return np.empty((0, 1))
-        if self._data_scaled is None:
-            self._data_scaled = self._scale_data(self._raw_data)
-        return self._data_scaled
+            return np.empty((0, ), dtype=self.dtype)
+        return self._scale_data(self._raw_data)
 
     @_property_builtin
     def raw_data(self):
-        """
-        The raw, unscaled data array.
+        """ If the TdmsFile was created by reading all data, this property
+        provides direct access to the numpy array of raw, unscaled data.
         For unscaled objects this is the same as the data property.
         """
         if len(self) > 0 and self._raw_data is None:
             raise RuntimeError("Channel data has not been read")
 
         if self._raw_data is None:
-            return np.empty((0, 1))
+            return np.empty((0, ), dtype=self._raw_data_dtype())
         if self._raw_data.scaler_data:
             if len(self._raw_data.scaler_data) == 1:
                 return next(v for v in self._raw_data.scaler_data.values())
@@ -526,27 +676,43 @@ class TdmsChannel(object):
 
     @_property_builtin
     def raw_scaler_data(self):
-        """ Raw DAQmx scaler data as a dictionary mapping from scale id to raw data arrays
+        """ If the TdmsFile was created by reading all data, this property
+        provides direct access to the numpy array of raw DAQmx scaler data
+        as a dictionary mapping from scale id to raw data arrays.
         """
         if len(self) > 0 and self._raw_data is None:
             raise RuntimeError("Channel data has not been read")
 
         return self._raw_data.scaler_data
 
+    def data_chunks(self):
+        """ A generator that streams chunks data for this channel from disk.
+        This method may only be used when the TDMS file was opened without reading all data immediately.
+
+        :rtype: Generator that yields :class:`ChannelDataChunk` objects
+        """
+        channel_offset = 0
+        for raw_data_chunk in self._tdms_file._read_channel_data_chunks(self):
+            yield ChannelDataChunk(self._tdms_file, self, raw_data_chunk, channel_offset)
+            channel_offset += len(raw_data_chunk)
+
     def read_data(self, offset=0, length=None, scaled=True):
-        """ Reads data for this channel from the TDMS file and returns it
+        """ Reads data for this channel from the TDMS file and returns it as a numpy array
 
-            This is for use when the TDMS file was opened without immediately reading all data,
-            otherwise the data attribute should be used.
+        Indexing into the channel with a slice should be preferred over using
+        this method, but this method is needed if you want to read raw, unscaled data.
 
-            :param offset: Initial position to read data from.
-            :param length: Number of values to attempt to read.
-                Fewer values will be returned if attempting to read beyond the end of the available data.
-            :param scaled: By default scaling will be applied to the returned data.
-                Set this parameter to False to return raw unscaled data.
-                For DAQmx data a dictionary of scaler id to raw scaler data will be returned.
+        :param offset: Initial position to read data from.
+        :param length: Number of values to attempt to read.
+            Fewer values will be returned if attempting to read beyond the end of the available data.
+        :param scaled: By default scaling will be applied to the returned data.
+            Set this parameter to False to return raw unscaled data.
+            For DAQmx data a dictionary of scaler id to raw scaler data will be returned.
         """
         raw_data = self._tdms_file._read_channel_data(self, offset, length)
+        if raw_data is None:
+            dtype = self.dtype if scaled else self._raw_data_dtype()
+            return np.empty((0,), dtype=dtype)
         if scaled:
             return self._scale_data(raw_data)
         else:
@@ -631,8 +797,73 @@ class TdmsChannel(object):
 
         return pandas_export.from_channel(self, time_index, absolute_time, scaled_data)
 
+    def _read_data_values(self):
+        for chunk in self.data_chunks():
+            for value in chunk:
+                yield value
+
+    def _read_slice(self, start, stop, step):
+        if step == 0:
+            raise ValueError("Step size cannot be zero")
+
+        # Replace None values with defaults
+        step = 1 if step is None else step
+        if start is None:
+            start = 0 if step > 0 else -1
+        if stop is None:
+            stop = self._length if step > 0 else -1 - self._length
+
+        # Handle negative indices
+        if start < 0:
+            start = self._length + start
+        if stop < 0:
+            stop = self._length + stop
+
+        # Check for empty ranges
+        if stop == start:
+            return np.empty((0, ), dtype=self.dtype)
+        if step > 0 and (stop < start or start >= self._length or stop < 0):
+            return np.empty((0,), dtype=self.dtype)
+        if step < 0 and (stop > start or stop >= self._length or start < 0):
+            return np.empty((0,), dtype=self.dtype)
+
+        # Trim values outside bounds
+        if start < 0:
+            start = 0
+        if start >= self._length:
+            start = self._length - 1
+        if stop > self._length:
+            stop = self._length
+        if stop < -1:
+            stop = -1
+
+        # Read data and handle step size
+        if step > 0:
+            read_data = self.read_data(start, stop - start)
+            return read_data[::step] if step > 1 else read_data
+        else:
+            read_data = self.read_data(stop + 1, start - stop)
+            return read_data[::step]
+
+    def _read_at_index(self, index):
+        if index < 0 or index >= self._length:
+            raise IndexError("Index {0} is outside of the channel bounds [0, {1}]".format(index, self._length - 1))
+
+        if self._cached_chunk is not None:
+            # Check if we've already read and cached the chunk containing this index
+            bounds = self._cached_chunk_bounds
+            if bounds[0] <= index < bounds[1]:
+                return self._cached_chunk[index - bounds[0]]
+
+        chunk, chunk_offset = self._tdms_file._read_channel_data_chunk_for_index(self, index)
+        scaled_chunk = self._scale_data(chunk)
+        self._cached_chunk = scaled_chunk
+        self._cached_chunk_bounds = (chunk_offset, chunk_offset + len(scaled_chunk))
+
+        return scaled_chunk[index - chunk_offset]
+
     def _scale_data(self, raw_data):
-        scale = self._get_scaling()
+        scale = self._scaling
         if scale is not None:
             return scale.scale(raw_data)
         elif raw_data.scaler_data:
@@ -640,7 +871,8 @@ class TdmsChannel(object):
         else:
             return raw_data.data
 
-    def _get_scaling(self):
+    @cached_property
+    def _scaling(self):
         group_properties = self._tdms_file[self._path.group].properties
         file_properties = self._tdms_file.properties
         return scaling.get_scaling(
@@ -682,7 +914,7 @@ class TdmsChannel(object):
 
     @_property_builtin
     def has_data(self):
-        """Deprecated"""
+        """(Deprecated)"""
         _deprecated("TdmsChannel.has_data", "This always returns True")
         return True
 
@@ -693,6 +925,114 @@ class TdmsChannel(object):
         return self._length
 
 
+class DataChunk(object):
+    """ A chunk of data in a TDMS file
+
+    Can be indexed by group name to get the data for a group in this channel,
+    which can then be indexed by channel name to get the data for a channel in this chunk.
+    For example::
+
+        group_chunk = data_chunk[group_name]
+        channel_chunk = group_chunk[channel_name]
+    """
+    def __init__(self, tdms_file, raw_data_chunk, channel_offsets):
+        self._groups = OrderedDict(
+            (group.name, GroupDataChunk(tdms_file, group, raw_data_chunk, channel_offsets))
+            for group in tdms_file.groups())
+
+    def __getitem__(self, group_name):
+        """ Get a chunk of data for a group
+        """
+        return self._groups[group_name]
+
+    def groups(self):
+        """ Returns chunks of data for all groups
+
+        :rtype: List of :class:`GroupDataChunk`
+        """
+        return list(self._groups.values())
+
+
+class GroupDataChunk(object):
+    """ A chunk of data for a group in a TDMS file
+
+    Can be indexed by channel name to get the data for a channel in this chunk.
+    For example::
+
+        channel_chunk = group_chunk[channel_name]
+
+    :ivar ~.name: Name of the group
+    """
+    def __init__(self, tdms_file, group, raw_data_chunk, channel_offsets):
+        self.name = group.name
+        self._channels = OrderedDict(
+            (channel.name, ChannelDataChunk(
+                tdms_file,
+                channel,
+                raw_data_chunk.channel_data.get(channel.path, RawChannelDataChunk.empty()),
+                channel_offsets[channel.path]))
+            for channel in group.channels())
+
+    def __getitem__(self, channel_name):
+        """ Get a chunk of data for a channel in this group
+        """
+        return self._channels[channel_name]
+
+    def channels(self):
+        """ Returns chunks of channel data for all channels in this group
+
+        :rtype: List of :class:`ChannelDataChunk`
+        """
+        return list(self._channels.values())
+
+
+class ChannelDataChunk(object):
+    """ A chunk of data for a channel in a TDMS file
+
+    Is an array-like object that supports indexing to access data, for example::
+
+        chunk_length = len(channel_data_chunk)
+        chunk_data = channel_data_chunk[:]
+
+    :ivar ~.name: Name of the channel
+    :ivar ~.offset: Starting index of this chunk of data in the entire channel
+    """
+    def __init__(self, tdms_file, channel, raw_data_chunk, offset):
+        self._path = channel._path
+        self._tdms_file = tdms_file
+        self._channel = channel
+        self.name = channel.name
+        self.offset = offset
+        self._raw_data = raw_data_chunk
+
+    def __len__(self):
+        """ Returns the number of values in this chunk
+        """
+        return len(self._raw_data)
+
+    def __getitem__(self, index):
+        """ Get a value or slice of values from this chunk
+        """
+        return self._data()[index]
+
+    def __iter__(self):
+        """ Iterate over values in this chunk
+        """
+        return iter(self._data())
+
+    def _data(self):
+        if self._raw_data.data is None and self._raw_data.scaler_data is None:
+            return np.empty((0, ), dtype=self._channel.dtype)
+
+        scale = self._channel._scaling
+        if scale is not None:
+            return scale.scale(self._raw_data)
+        elif self._raw_data.scaler_data:
+            raise ValueError("Missing scaling information for DAQmx data")
+        else:
+            return self._raw_data.data
+
+
 class RootObject(object):
     def __init__(self, properties):
         self.properties = properties
