@@ -13,11 +13,16 @@ from nptdms.test.util import (
     channel_metadata_with_repeated_structure)
 
 
+TDS_TYPE_INT8 = 1
+TDS_TYPE_INT16 = 2
 TDS_TYPE_INT32 = 3
+TDS_TYPE_BOOL = 0x21
 TDS_TYPE_COMPLEX64 = 0x08000c
 TDS_TYPE_COMPLEX128 = 0x10000d
 TDS_TYPE_FLOAT32 = 9
 TDS_TYPE_FLOAT64 = 10
+TDS_TYPE_FLOAT32_WITH_UNIT = 0x19
+TDS_TYPE_FLOAT64_WITH_UNIT = 0x1A
 
 
 _scenarios = []
@@ -89,6 +94,28 @@ def single_segment_with_two_channels_interleaved():
     return test_file, expected_data
 
 
+@scenario
+def single_segment_with_interleaved_data_of_different_width():
+    test_file = GeneratedFile()
+    test_file.add_segment(
+        ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData"),
+        segment_objects_metadata(
+            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT8, 2),
+            channel_metadata("/'group'/'channel2'", TDS_TYPE_INT16, 2),
+            channel_metadata("/'group'/'channel3'", TDS_TYPE_INT32, 2),
+        ),
+        "01" "01 00" "01 00 00 00"
+        "02" "02 00" "02 00 00 00"
+        "03" "03 00" "03 00 00 00"
+    )
+    expected_data = {
+        ('group', 'channel1'): np.array([1, 2, 3], dtype=np.int8),
+        ('group', 'channel2'): np.array([1, 2, 3], dtype=np.int16),
+        ('group', 'channel3'): np.array([1, 2, 3], dtype=np.int32),
+    }
+    return test_file, expected_data
+
+
 @scenario
 def no_metadata_segment():
     """ Add a segment with two channels, then a second
@@ -372,22 +399,22 @@ def chunked_segment():
             channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
             channel_metadata("/'group'/'channel2'", TDS_TYPE_INT32, 2),
         ),
-        "01 00 00 00" "02 00 00 00"
-        "03 00 00 00" "04 00 00 00"
-        "05 00 00 00" "06 00 00 00"
-        "07 00 00 00" "08 00 00 00"
+        "00 00 00 00" "01 00 00 00"
+        "0A 00 00 00" "0B 00 00 00"
+        "02 00 00 00" "03 00 00 00"
+        "0C 00 00 00" "0D 00 00 00"
     )
     test_file.add_segment(
         ("kTocRawData", ),
         "",
-        "07 00 00 00" "08 00 00 00"
-        "05 00 00 00" "06 00 00 00"
-        "03 00 00 00" "04 00 00 00"
-        "01 00 00 00" "02 00 00 00"
+        "04 00 00 00" "05 00 00 00"
+        "0E 00 00 00" "0F 00 00 00"
+        "06 00 00 00" "07 00 00 00"
+        "10 00 00 00" "11 00 00 00"
     )
     expected_data = {
-        ('group', 'channel1'): np.array([1, 2, 5, 6, 7, 8, 3, 4], dtype=np.int32),
-        ('group', 'channel2'): np.array([3, 4, 7, 8, 5, 6, 1, 2], dtype=np.int32),
+        ('group', 'channel1'): np.array([0, 1, 2, 3, 4, 5, 6, 7], dtype=np.int32),
+        ('group', 'channel2'): np.array([10, 11, 12, 13, 14, 15, 16, 17], dtype=np.int32),
     }
     return test_file, expected_data
 
@@ -559,6 +586,26 @@ def incomplete_last_row_of_interleaved_data():
     return test_file, expected_data
 
 
+@scenario
+def bool_data():
+    """ Test reading a file with boolean valued data
+    """
+    expected_channel_data = np.array([False, True, False, True], dtype=np.dtype('bool8'))
+
+    test_file = GeneratedFile()
+    test_file.add_segment(
+        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
+        segment_objects_metadata(
+            channel_metadata("/'group'/'bool_channel'", TDS_TYPE_BOOL, 2),
+        ),
+        "00 01 00 01"
+    )
+    expected_data = {
+        ('group', 'bool_channel'): expected_channel_data,
+    }
+    return test_file, expected_data
+
+
 @scenario
 def float_data():
     """ Test reading a file with float valued data
@@ -591,6 +638,40 @@ def float_data():
     return test_file, expected_data
 
 
+@scenario
+def float_data_with_unit():
+    """ Test reading a file with float valued data with units
+
+        These are the same as normal floating point data but have a 'unit_string' property
+    """
+    single_arr = np.array([0.123, 0.234, 0.345, 0.456], dtype=np.float32)
+    double_arr = np.array([0.987, 0.876, 0.765, 0.654], dtype=np.double)
+    data = ""
+    for num in single_arr[0:2]:
+        data += hexlify_value("<f", num)
+    for num in double_arr[0:2]:
+        data += hexlify_value("<d", num)
+    for num in single_arr[2:4]:
+        data += hexlify_value("<f", num)
+    for num in double_arr[2:4]:
+        data += hexlify_value("<d", num)
+
+    test_file = GeneratedFile()
+    test_file.add_segment(
+        ("kTocMetaData", "kTocRawData", "kTocNewObjList"),
+        segment_objects_metadata(
+            channel_metadata("/'group'/'single_channel'", TDS_TYPE_FLOAT32_WITH_UNIT, 2),
+            channel_metadata("/'group'/'double_channel'", TDS_TYPE_FLOAT64_WITH_UNIT, 2),
+        ),
+        data
+    )
+    expected_data = {
+        ('group', 'single_channel'): single_arr,
+        ('group', 'double_channel'): double_arr,
+    }
+    return test_file, expected_data
+
+
 @scenario
 def complex_data():
     """ Test reading a file with complex valued data
@@ -639,10 +720,10 @@ def scaled_data():
         segment_objects_metadata(
             channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2, properties),
         ),
-        "01 00 00 00" "02 00 00 00"
+        "01 00 00 00" "02 00 00 00" "03 00 00 00" "04 00 00 00"
     )
     expected_data = {
-        ('group', 'channel1'): np.array([12, 14], dtype=np.float64),
+        ('group', 'channel1'): np.array([12, 14, 16, 18], dtype=np.float64),
     }
     return test_file, expected_data
 
@@ -695,7 +776,7 @@ def timestamp_data():
 
     test_file = GeneratedFile()
     toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList")
-    test_file.add_segment(toc, metadata, _timestamp_data(times))
+    test_file.add_segment(toc, metadata, timestamp_data_chunk(times))
 
     expected_data = {
         ('Group', 'TimeChannel1'): np.array([times[0], times[1]]),
@@ -753,7 +834,7 @@ def interleaved_timestamp_data():
 
     test_file = GeneratedFile()
     toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData")
-    test_file.add_segment(toc, metadata, _timestamp_data(times))
+    test_file.add_segment(toc, metadata, timestamp_data_chunk(times))
 
     expected_data = {
         ('Group', 'TimeChannel1'): np.array([times[0], times[2]]),
@@ -763,7 +844,137 @@ def interleaved_timestamp_data():
     return test_file, expected_data
 
 
-def _timestamp_data(times):
+@scenario
+def interleaved_timestamp_and_numpy_data():
+    """Test reading timestamp data interleaved with a standard numpy data type
+    """
+
+    times = [
+        np.datetime64('2012-08-23T00:00:00.123', 'us'),
+        np.datetime64('2012-08-23T01:02:03.456', 'us'),
+    ]
+
+    metadata = (
+        # Number of objects
+        "02 00 00 00"
+        # Length of the object path
+        "16 00 00 00")
+    metadata += string_hexlify("/'Group'/'TimeChannel'")
+    metadata += (
+        # Length of index information
+        "14 00 00 00"
+        # Raw data data type
+        "44 00 00 00"
+        # Dimension
+        "01 00 00 00"
+        # Number of raw data values
+        "02 00 00 00"
+        "00 00 00 00"
+        # Number of properties (0)
+        "00 00 00 00")
+    metadata += (
+        "15 00 00 00")
+    metadata += string_hexlify("/'Group'/'IntChannel'")
+    metadata += (
+        # Length of index information
+        "14 00 00 00"
+        # Raw data data type
+        "03 00 00 00"
+        # Dimension
+        "01 00 00 00"
+        # Number of raw data values
+        "02 00 00 00"
+        "00 00 00 00"
+        # Number of properties (0)
+        "00 00 00 00")
+
+    data = (timestamp_data_chunk([times[0]]) +
+            "01 00 00 00" +
+            timestamp_data_chunk([times[1]]) +
+            "02 00 00 00")
+
+    test_file = GeneratedFile()
+    toc = ("kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocInterleavedData")
+    test_file.add_segment(toc, metadata, data)
+
+    expected_data = {
+        ('Group', 'TimeChannel'): np.array([times[0], times[1]]),
+        ('Group', 'IntChannel'): np.array([1, 2], dtype=np.dtype('int32')),
+    }
+
+    return test_file, expected_data
+
+
+@scenario
+def segment_without_data():
+    test_file = GeneratedFile()
+    test_file.add_segment(
+        ("kTocMetaData", "kTocNewObjList"),
+        segment_objects_metadata(
+            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
+        ),
+        ""
+    )
+    test_file.add_segment(
+        ("kTocMetaData", "kTocNewObjList", "kTocRawData"),
+        segment_objects_metadata(
+            channel_metadata_with_repeated_structure("/'group'/'channel1'"),
+        ),
+        "01 00 00 00" "02 00 00 00"
+    )
+    test_file.add_segment(
+        ("kTocMetaData", "kTocNewObjList"),
+        segment_objects_metadata(
+            channel_metadata_with_repeated_structure("/'group'/'channel1'"),
+        ),
+        ""
+    )
+    test_file.add_segment(
+        ("kTocMetaData", "kTocNewObjList", "kTocRawData"),
+        segment_objects_metadata(
+            channel_metadata_with_repeated_structure("/'group'/'channel1'"),
+        ),
+        "03 00 00 00" "04 00 00 00"
+    )
+    expected_data = {
+        ('group', 'channel1'): np.array([1, 2, 3, 4], dtype=np.int32),
+    }
+    return test_file, expected_data
+
+
+@scenario
+def channel_without_data():
+    test_file = GeneratedFile()
+    test_file.add_segment(
+        ("kTocMetaData", "kTocNewObjList"),
+        segment_objects_metadata(
+            channel_metadata("/'group'/'channel1'", TDS_TYPE_INT32, 2),
+        ),
+        ""
+    )
+    expected_data = {
+        ('group', 'channel1'): np.array([], dtype=np.dtype('int32')),
+    }
+    return test_file, expected_data
+
+
+@scenario
+def channel_without_data_or_data_type():
+    test_file = GeneratedFile()
+    test_file.add_segment(
+        ("kTocMetaData", "kTocNewObjList"),
+        segment_objects_metadata(
+            channel_metadata_with_no_data("/'group'/'channel1'"),
+        ),
+        ""
+    )
+    expected_data = {
+        ('group', 'channel1'): np.array([], dtype=np.dtype('void')),
+    }
+    return test_file, expected_data
+
+
+def timestamp_data_chunk(times):
     epoch = np.datetime64('1904-01-01T00:00:00')
 
     def total_seconds(td):
