@@ -1,9 +1,12 @@
 """Test reading of TDMS files with DAQmx data
 """
 
+from collections import defaultdict
+import logging
 import numpy as np
 
 from nptdms import TdmsFile
+from nptdms.log import log_manager
 from nptdms.test.util import (
     GeneratedFile, hexlify_value, string_hexlify, segment_objects_metadata, hex_properties)
 
@@ -400,6 +403,58 @@ def test_multiple_raw_data_buffers_with_scalers_split_across_buffers():
     np.testing.assert_array_equal(scaler_data_4, [10, 12, 14, 16])
 
 
+def test_digital_line_scaler_data():
+    """ Test loading a DAQmx file with a single channel of U8 digital line scaler data
+    """
+
+    scaler_metadata = daqmx_scaler_metadata(0, 0, 2, digital_line_scaler=True)
+    metadata = segment_objects_metadata(
+        root_metadata(),
+        group_metadata(),
+        daqmx_channel_metadata("Channel1", 4, [4], [scaler_metadata], digital_line_scaler=True))
+    data = (
+        "00 00 00 00"
+        "00 00 01 00"
+        "00 00 00 00"
+        "00 00 01 00"
+    )
+
+    test_file = GeneratedFile()
+    test_file.add_segment(segment_toc(), metadata, data)
+    tdms_data = test_file.load()
+
+    data = tdms_data["Group"]["Channel1"].raw_data
+
+    assert data.dtype == np.uint8
+    np.testing.assert_array_equal(data, [0, 1, 0, 1])
+
+
+def test_digital_line_scaler_data_uses_first_bit_of_bytes():
+    """ Test DAQmx digital line scaler data only uses the first bit in each byte to represent a 1 or 0 value
+    """
+
+    scaler_metadata = daqmx_scaler_metadata(0, 0, 2, digital_line_scaler=True)
+    metadata = segment_objects_metadata(
+        root_metadata(),
+        group_metadata(),
+        daqmx_channel_metadata("Channel1", 4, [4], [scaler_metadata], digital_line_scaler=True))
+    data = (
+        "00 00 00 00"
+        "00 00 01 00"
+        "00 00 02 00"
+        "00 00 03 00"
+    )
+
+    test_file = GeneratedFile()
+    test_file.add_segment(segment_toc(), metadata, data)
+    tdms_data = test_file.load()
+
+    data = tdms_data["Group"]["Channel1"].raw_data
+
+    assert data.dtype == np.uint8
+    np.testing.assert_array_equal(data, [0, 1, 0, 1])
+
+
 def test_lazily_reading_channel():
     """ Test loading channels individually from a DAQmx file
     """
@@ -525,6 +580,79 @@ def test_lazily_reading_a_subset_of_raw_channel_data():
             np.testing.assert_array_equal(data_2[0], [18, 19])
 
 
+def test_stream_data_chunks():
+    """Test streaming chunks of DAQmx data from a TDMS file
+    """
+    properties = {
+        "NI_Number_Of_Scales": (3, "01 00 00 00"),
+    }
+    scaler_1 = daqmx_scaler_metadata(0, 3, 0)
+    scaler_2 = daqmx_scaler_metadata(0, 3, 2)
+    metadata = segment_objects_metadata(
+        root_metadata(),
+        group_metadata(),
+        daqmx_channel_metadata("Channel1", 4, [4], [scaler_1], properties),
+        daqmx_channel_metadata("Channel2", 4, [4], [scaler_2], properties))
+    data = (
+        # Data for segment
+        "01 00" "11 00"
+        "02 00" "12 00"
+        "03 00" "13 00"
+        "04 00" "14 00"
+        "05 00" "15 00"
+        "06 00" "16 00"
+        "07 00" "17 00"
+        "08 00" "18 00"
+    )
+
+    test_file = GeneratedFile()
+    test_file.add_segment(segment_toc(), metadata, data)
+    data_arrays = defaultdict(list)
+    with test_file.get_tempfile() as temp_file:
+        with TdmsFile.open(temp_file.file) as tdms_file:
+            for chunk in tdms_file.data_chunks():
+                for group in chunk.groups():
+                    for channel in group.channels():
+                        key = (group.name, channel.name)
+                        assert channel.offset == len(data_arrays[key])
+                        data_arrays[key].extend(channel[:])
+
+    expected_channel_data = {
+        ("Group", "Channel1"): [1, 2, 3, 4, 5, 6, 7, 8],
+        ("Group", "Channel2"): [17, 18, 19, 20, 21, 22, 23, 24],
+    }
+    for ((group, channel), expected_data) in expected_channel_data.items():
+        actual_data = data_arrays[(group, channel)]
+        np.testing.assert_equal(actual_data, expected_data)
+
+
+def test_daqmx_debug_logging(caplog):
+    """ Test loading a DAQmx file with debug logging enabled
+    """
+    scaler_metadata = daqmx_scaler_metadata(0, 3, 0)
+    metadata = segment_objects_metadata(
+        root_metadata(),
+        group_metadata(),
+        daqmx_channel_metadata("Channel1", 4, [2], [scaler_metadata]))
+    data = (
+        "01 00"
+        "02 00"
+        "FF FF"
+        "FE FF"
+    )
+
+    test_file = GeneratedFile()
+    test_file.add_segment(segment_toc(), metadata, data)
+
+    log_manager.set_level(logging.DEBUG)
+    _ = test_file.load()
+
+    assert "Reading metadata for object /'Group'/'Channel1' with index header 0x00001269" in caplog.text
+    assert "scaler_type=4713" in caplog.text
+    assert "scale_id=0" in caplog.text
+    assert "data_type=Int16" in caplog.text
+
+
 def segment_toc():
     return (
         "kTocMetaData", "kTocRawData", "kTocNewObjList", "kTocDAQmxRawData")
@@ -555,7 +683,7 @@ def group_metadata():
         "00 00 00 00")
 
 
-def daqmx_scaler_metadata(scale_id, type_id, byte_offset, raw_buffer_index=0):
+def daqmx_scaler_metadata(scale_id, type_id, byte_offset, raw_buffer_index=0, digital_line_scaler=False):
     return (
         # DAQmx data type (type ids don't match TDMS types)
         hexlify_value("<I", type_id) +
@@ -564,14 +692,14 @@ def daqmx_scaler_metadata(scale_id, type_id, byte_offset, raw_buffer_index=0):
         # Raw byte offset
         hexlify_value("<I", byte_offset) +
         # Sample format bitmap (don't know what this is for...)
-        "00 00 00 00" +
+        ("00" if digital_line_scaler else "00 00 00 00") +
         # Scale ID
         hexlify_value("<I", scale_id))
 
 
 def daqmx_channel_metadata(
         channel_name, num_values,
-        raw_data_widths, scaler_metadata, properties=None):
+        raw_data_widths, scaler_metadata, properties=None, digital_line_scaler=False):
     path = "/'Group'/'" + channel_name + "'"
     return (
         # Length of the object path
@@ -579,7 +707,7 @@ def daqmx_channel_metadata(
         # Object path
         string_hexlify(path) +
         # Raw data index (DAQmx)
-        "69 12 00 00"
+        ("6A 12 00 00" if digital_line_scaler else "69 12 00 00") +
         # Data type (DAQmx)
         "FF FF FF FF"
         # Array  dimension
